#include "Freelook.h"
#include "MatrixStack.h"
#include <iostream>

Freelook::Freelook() :
	aspect(1.0f),
	fovy(45.0f),
	znear(0.1f),
	zfar(1000.0f),
	rotations(0.0, 0.0),
	translations(0.0f, 0.0f, -5.0f),
	rfactor(0.002f),
	tfactor(0.05f),
	sfactor(0.005f),
	pos(0.0f, 0.0f, -5.0f),
	yaw(0.0f),
	pitch(0.0f),
	move(0.0f, 0.0f, -1.0f),
	up(0.0f, 1.0f, 0.0f)

{
}

Freelook::~Freelook()
{
}

void Freelook::mouseClicked(float x, float y)
{
	mousePrev << x, y;
		state = Freelook::ROTATE;
}


void Freelook::mouseMoved(float x, float y)
{
	Eigen::Vector2f mouseCurr(x, y);
	Eigen::Vector2f dv = mouseCurr - mousePrev;

	dv = dv * rfactor;

	yaw = yaw + dv.x();
	pitch = pitch + dv.y(); 
	
	if (pitch > 60.0f)
		pitch = 60.0f;
	if (pitch < -60.0f)
		pitch = -60.0f;

	Eigen::Vector3f view;
	view.x() = cos(yaw) * cos(pitch);
	view.y() = sin(-pitch);
	view.z() = sin(yaw) * cos(pitch);
	view.normalize();
	move = view;
	mousePrev = mouseCurr;
}

void Freelook::keyPressed(bool keyToggles[256])
{
	Eigen::Vector3f dir = move.cross(up);
	dir.normalize();

	if (keyToggles[(unsigned int)'a'] && !(pos.x() <= -0.9f))
		pos = pos - dir * tfactor;

	if (keyToggles[(unsigned int)'d'] && !(pos.x() >= 0.9f))
		pos = pos + dir * tfactor;

	if (keyToggles[(unsigned int)'s'] && (pos.x() >= 0.0f))
		pos = pos - dir * tfactor;

	if (keyToggles[(unsigned int)'s'] && (pos.x() <= 0.0f))
		pos = pos + dir * tfactor;


}

void Freelook::applyProjectionMatrix(std::shared_ptr<MatrixStack> P) const
{
	// Modify provided MatrixStack
	P->perspective(fovy, aspect, znear, zfar);
}

void Freelook::applyViewMatrix(std::shared_ptr<MatrixStack> MV) const
{
	MV->lookAt(pos,pos + move, Eigen::Vector3f(0,1,0));
}
