#pragma  once
#ifndef __Freelook__
#define __Freelook__

#include <memory>

#define EIGEN_DONT_ALIGN_STATICALLY
#include <Eigen/Dense>

class MatrixStack;

class Freelook
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	
	enum {
		ROTATE = 0,
		TRANSLATE,
		SCALE
	};
	
	Freelook();
	virtual ~Freelook();
	void setInitDistance(float z) { translations(2) = -std::abs(z); }
	void setAspect(float a) { aspect = a; };
	void setRotationFactor(float f) { rfactor = f; };
	void setTranslationFactor(float f) { tfactor = f; };
	void setScaleFactor(float f) { sfactor = f; };
	void mouseClicked(float x, float y);
	void keyPressed(bool keyToggles[256]);
	void mouseMoved(float x, float y);
	void applyProjectionMatrix(std::shared_ptr<MatrixStack> P) const;
	void applyViewMatrix(std::shared_ptr<MatrixStack> MV) const;
	Eigen::Vector3f pos;
	
private:
	float aspect;
	float fovy;
	float znear;
	float zfar;
	Eigen::Vector2f rotations;
	Eigen::Vector3f translations;
	Eigen::Vector2f mousePrev;
	int state;
	float rfactor;
	float tfactor;
	float sfactor;
	float yaw;
	float pitch;
	Eigen::Vector3f move;
	Eigen::Vector3f up;
};

#endif
