#include <cassert>
#include <cstring>
#define _USE_MATH_DEFINES
#include <cmath>
#include <iostream>

#define GLEW_STATIC
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#define EIGEN_DONT_ALIGN_STATICALLY
#include <Eigen/Dense>

#include "Camera.h"
#include "Freelook.h"
#include "GLSL.h"
#include "MatrixStack.h"
#include "Program.h"
#include "Particle.h"
#include "Shape.h"
#include "Light.h"
#include "Texture.h"
#include "Component.h"
#include "Target.h"

using namespace std;
using namespace Eigen;

GLFWwindow *window; // Main application window
string RESOURCE_DIR = "./"; // Where the resources are loaded from

shared_ptr<Camera> camera;
shared_ptr<Freelook> fl;
shared_ptr<Program> prog;
shared_ptr<Program> text;
shared_ptr<Program> marb;
vector< shared_ptr< Particle> >  marble;
Vector3f grav;
float t, h;

shared_ptr<Component> b1;
shared_ptr<Component> flr;
shared_ptr<Component> back;
shared_ptr<Component> bld;
shared_ptr<Component> c1;
shared_ptr<Component> lgt;
shared_ptr<Target> targ;

shared_ptr<Shape> bunny;
shared_ptr<Shape> square;
shared_ptr<Shape> sphere;
shared_ptr<Shape> cube;
shared_ptr<Texture> tex;
shared_ptr<Texture> stone;
shared_ptr<Texture> bg;
shared_ptr<Texture> texture0;

bool keyToggles[256] = {false}; // only for English keyboards!

const Vector3f xAxis(1.0, 0, 0);
const Vector3f yAxis(0, 1.0, 0);
const Vector3f zAxis(0, 0, 1.0);
const Vector3f move(0, 0, -0.01f);

int material = 0;
int xBut = 0;
int yBut = 0;
int shad = 0;
shared_ptr<Light> light0; 

// This function is called when a GLFW error occurs
static void error_callback(int error, const char *description)
{
	cerr << description << endl;
}

// If the window is resized, capture the new size and reset the viewport
static void resize_callback(GLFWwindow *window, int width, int height)
{
	glViewport(0, 0, width, height);
}

// This function is called when the mouse moves
static void cursor_position_callback(GLFWwindow* window, double xmouse, double ymouse)
{
}



// This function is called when the mouse is clicked
static void mouse_button_callback(GLFWwindow *window, int button, int action, int mods)
{
	// Get the current mouse position.
	double xmouse, ymouse;
	glfwGetCursorPos(window, &xmouse, &ymouse);
	// Get current window size.
	int width, height;
	glfwGetWindowSize(window, &width, &height);
	if(action == GLFW_PRESS) {
		marble[0]->rebirth(5.0f, (float)(xmouse / width * 2) - 1, (float)(ymouse/height * 2) - 2, fl->pos);
	}
}

// This function is called when a key is pressed
static void key_callback(GLFWwindow *window, int key, int scancode, int action, int mods)
{
	if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
		glfwSetWindowShouldClose(window, GL_TRUE);
	}	
	if (key == 'm' || key == 'M') {		//rotate through the materials
		material = (material + 1) % 3;
	}
	
	if ((key == 'p' || key == 'P') && action == GLFW_PRESS)
	{
		keyToggles[(unsigned int)'p'] = !keyToggles[(unsigned int)'p'];
	}

	if ((key == 'a' || key == 'A') && action == GLFW_PRESS)
	{
		keyToggles[(unsigned int)'a'] = true;
	}
	else if ((key == 'a' || key == 'A') && action == GLFW_RELEASE)
	{
		keyToggles[(unsigned int)'a'] = false;
	}

	if ((key == 's' || key == 'S') && action == GLFW_PRESS)
	{
		keyToggles[(unsigned int)'s'] = true;
	}
	else if ((key == 's' || key == 'S') && action == GLFW_RELEASE)
	{
		keyToggles[(unsigned int)'s'] = false;
	}

	if ((key == 'd' || key == 'D') && action == GLFW_PRESS)
	{
		keyToggles[(unsigned int)'d'] = true;
	}
	else if ((key == 'd' || key == 'D') && action == GLFW_RELEASE)
	{
		keyToggles[(unsigned int)'d'] = false;
	}    
	if (key != 'w' && key != 'a' && key != 's' && key != 'd')
		keyToggles[key] = !keyToggles[key];
}

static void char_callback(GLFWwindow *window, unsigned int key)
{

}


// This function is called once to initialize the scene and OpenGL
static void init()
{
	// Initialize time.
	glfwSetTime(0.0);
	
	// Set background color.
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	// Enable z-buffer test.
	glEnable(GL_DEPTH_TEST);
	// Enable setting gl_PointSize from vertex shader
	glEnable(GL_PROGRAM_POINT_SIZE);
	// Enable quad creation from sprite
	glEnable(GL_POINT_SPRITE);

	//frag shader setup
	prog = make_shared<Program>();
	prog->setShaderNames(RESOURCE_DIR + "vert.glsl", RESOURCE_DIR + "frag.glsl");
	prog->setVerbose(false);
	prog->init();
	prog->addAttribute("aPos");
	prog->addAttribute("aNor");
	prog->addUniform("MV");
	prog->addUniform("P");

	//light and material setup
	prog->addUniform("lightPos");
	prog->addUniform("ka");
	prog->addUniform("kd");
	prog->addUniform("ks");
	prog->addUniform("s");

	//texture setup
	text = make_shared<Program>();
	text->setShaderNames(RESOURCE_DIR + "textVert.glsl", RESOURCE_DIR + "texture.glsl");
	text->setVerbose(true);
	text->init();	
	text->addAttribute("aPos");
	text->addAttribute("aNor");
	text->addAttribute("aTex");
	text->addUniform("P");
	text->addUniform("MV");
	text->addUniform("val");
	text->addUniform("tex");
	text->addUniform("lightPos");

	//marble setup
	marb = make_shared<Program>();
	marb->setShaderNames(RESOURCE_DIR + "marbV.glsl", RESOURCE_DIR + "marbF.glsl");
	marb->setVerbose(true);
	marb->init();
	marb->addAttribute("aPos");
	marb->addAttribute("aAlp");
	marb->addAttribute("aCol");
	marb->addAttribute("aSca");
	marb->addUniform("P");
	marb->addUniform("MV");
	marb->addUniform("screenSize");
	marb->addUniform("texture0");

	//camera setup
	fl = make_shared<Freelook>();
	fl->setInitDistance(5.0f);

	//bunny setup
	bunny = make_shared<Shape>();
	bunny->loadMesh(RESOURCE_DIR + "bunny.obj");
	bunny->fitToUnitBox();
	bunny->init();

	//bunny 1
		//b1 transform
		b1 = make_shared<Component>();
		b1->setShape(bunny);
		const Vector3f b1Trans(2.0f, -0.5f, 1.0f);
		b1->setTransformMesh(b1Trans);
		const Vector3f b1Rot(0.0f, 0.0f, 0.0f);
		b1->setAngle(b1Rot);
		const Vector3f b1Scale(1.0f, 1.0f, 1.0f);
		b1->setScaleMesh(b1Scale);

		//b1 color
		const Vector3f b1Ka(0.3f, 0.3f, 0.4f);
		b1->setKa(b1Ka);
		const Vector3f b1Kd(0.4f, 0.4f, 0.4f);
		b1->setKd(b1Kd);
		const Vector3f b1Ks(0.05f, 0.05f, 0.05f);
		b1->setKs(b1Ks);
		b1->setS(200.0f);


	//square setup 
	square = make_shared<Shape>();
	square->loadMesh(RESOURCE_DIR + "square.obj");
	square->fitToUnitBox();
	square->init();

		//floor transform
		flr = make_shared<Component>();
		flr->setShape(square);
		const Vector3f fTrans(0.0f, -1.0f, 0.0f);
		flr->setTransformMesh(fTrans);
		const Vector3f fRot(-90.0f, 0.0f, 0.0f);
		flr->setAngle(fRot);
		const Vector3f fScale(3.0f, 200.0f, 0.01f);
		flr->setScaleMesh(fScale);

		//texture setup (floor)
		tex = make_shared<Texture>();
		tex->setFilename(RESOURCE_DIR + "floor.jpg");
		tex->init();
		tex->setUnit(0);
		tex->setWrapModes(GL_REPEAT, GL_REPEAT);

		//bg transform
		back = make_shared<Component>();
		back->setShape(square);
		const Vector3f bgTrans(0.0f, -1.0f, 0.0f);
		back->setTransformMesh(bgTrans);
		const Vector3f bgScale(200.0f, 200.0f, 0.01f);
		back->setScaleMesh(bgScale);

		//bg text
		bg = make_shared<Texture>();
		bg->setFilename(RESOURCE_DIR + "bg.jpg");
		bg->init();
		bg->setUnit(0);
		bg->setWrapModes(GL_REPEAT, GL_REPEAT);

		//target transform
		targ = make_shared<Target>();
		targ->setShape(square);
		const Vector3f tScale(0.2f, 0.1f, 0.001f);
		targ->setScaleMesh(tScale);

		////target color
		const Vector3f tKa(0.3f, 0.3f, 0.4f);
		targ->setKa(tKa);
		const Vector3f tKd(0.2f, 0.2f, 0.6f);
		targ->setKd(tKd);
		const Vector3f tKs(0.1f, 0.1f, 0.1f);
		targ->setKs(tKs);
		targ->setS(200.0f);


	//cube
	cube = make_shared<Shape>();
	cube->loadMesh(RESOURCE_DIR + "cube.obj");
	cube->fitToUnitBox();
	cube->init();
		//building transform
		bld = make_shared<Component>();
		bld->setShape(cube);
		const Vector3f bldRot(-90.0f, 0.0f, 0.0f);
		bld->setAngle(bldRot);
		const Vector3f bldScale(1.0f, 200.0f, 2.0f);
		bld->setScaleMesh(bldScale);

		//texture setup (floor)
		stone = make_shared<Texture>();
		stone->setFilename(RESOURCE_DIR + "stone.jpg");
		stone->init();
		stone->setUnit(0);
		stone->setWrapModes(GL_REPEAT, GL_REPEAT);








	//sphere setup
	const Vector3f lTrans(10.0f, 5.0f, 10.0f);
	light0 = make_shared<Light>();
	light0->setLightPos(lTrans);
		
		//light
		sphere = make_shared<Shape>();
		sphere->loadMesh(RESOURCE_DIR + "sphere.obj");
		sphere->fitToUnitBox();
		sphere->init();

		lgt = make_shared<Component>();
		lgt->setShape(sphere);
		lgt->setTransformMesh(lTrans);
		const Vector3f lScale(3.0f, 3.0f, 3.0f);
		lgt->setScaleMesh(lScale);		
		
		//light color
		const Vector3f lgtKa(1.0f, 0.8f, 0.0f);
		lgt->setKa(lgtKa);
		const Vector3f lgtKd(0.7f, 0.3f, 0.4f);
		lgt->setKd(lgtKd);
		const Vector3f lgtKs(0.05f, 0.05f, 0.05f);
		lgt->setKs(lgtKs);
		lgt->setS(50.0f);

		//marble setup			
		texture0 = make_shared<Texture>();
		texture0->setFilename(RESOURCE_DIR + "alpha.jpg");
		texture0->init();
		texture0->setUnit(0);
		texture0->setWrapModes(GL_REPEAT, GL_REPEAT);
		int n = 1;
		Particle::init(n);
		for (int i = 0; i < n; ++i) {
			auto p = make_shared<Particle>(i);
			marble.push_back(p);
			//p->rebirth(0.0f, keyToggles);
		}
		grav << 0.0f, -9.8f, 0.0f;
		t = 0.0f;
		h = 0.01f;

	GLSL::checkError(GET_FILE_LINE);
}

// This function is called every frame to draw the scene.
static void render()
{
	// Clear framebuffer.
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	if (keyToggles[(unsigned)'c']) {
		glEnable(GL_CULL_FACE);
	}
	else {
		glDisable(GL_CULL_FACE);
	}
	if (keyToggles[(unsigned)'l']) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else {
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	fl->keyPressed(keyToggles);

	// Get current frame buffer size.
	int width, height;
	glfwGetFramebufferSize(window, &width, &height);
	fl->setAspect((float)width / (float)height);

	// Matrix stacks
	auto P = make_shared<MatrixStack>();
	auto MV = make_shared<MatrixStack>();

	// Apply camera transforms
	P->pushMatrix();
		fl->applyProjectionMatrix(P);


		MV->pushMatrix();


			fl->applyViewMatrix(MV);

			MV->translate(0.0f, 0.0f, -110.0f); //-110
			text->bind();
				glUniformMatrix4fv(text->getUniform("P"), 1, GL_FALSE, P->topMatrix().data());
				glUniformMatrix4fv(text->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
				glUniform3f(text->getUniform("lightPos"), light0->getXPos(), light0->getYPos(), light0->getZPos());

				//floor
				MV->pushMatrix();
					tex->bind(text->getUniform("tex"));
					MV->translate(flr->getTransformMesh());
					MV->rotate(flr->getAngle().x(), xAxis);
					MV->rotate(flr->getAngle().y(), yAxis);
					MV->rotate(flr->getAngle().z(), zAxis);
					MV->scale(flr->getScaleMesh());
					glUniformMatrix4fv(text->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					flr->getShape()->draw(text);
					tex->unbind();
				MV->popMatrix();
				MV->pushMatrix();
					bg->bind(text->getUniform("tex"));
					MV->translate(back->getTransformMesh());
					MV->rotate(back->getAngle().x(), xAxis);
					MV->rotate(back->getAngle().y(), yAxis);
					MV->rotate(back->getAngle().z(), zAxis);
					MV->scale(back->getScaleMesh());
					glUniformMatrix4fv(text->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					back->getShape()->draw(text);
					bg->unbind();
				MV->popMatrix();
				MV->pushMatrix();
					stone->bind(text->getUniform("tex"));
					MV->translate(Vector3f(-2.0f, 0.0f, 0.0f));
					MV->rotate(bld->getAngle().x(), xAxis);
					MV->rotate(bld->getAngle().y(), yAxis);
					MV->rotate(bld->getAngle().z(), zAxis);
					MV->scale(bld->getScaleMesh());
					glUniformMatrix4fv(text->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					bld->getShape()->draw(text);
					stone->unbind();
				MV->popMatrix();				
				MV->pushMatrix();
					stone->bind(text->getUniform("tex"));
					MV->translate(Vector3f(2.0f, 0.0f, 0.0f));
					MV->rotate(bld->getAngle().x(), xAxis);
					MV->rotate(bld->getAngle().y(), yAxis);
					MV->rotate(bld->getAngle().z(), zAxis);
					MV->scale(bld->getScaleMesh());
					glUniformMatrix4fv(text->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					bld->getShape()->draw(text);
					stone->unbind();
				MV->popMatrix();	

			//frag shader

			prog->bind();

				glUniformMatrix4fv(prog->getUniform("P"), 1, GL_FALSE, P->topMatrix().data());
				glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
				glUniform3f(prog->getUniform("lightPos"), -light0->getXPos(), light0->getYPos(), light0->getZPos());

				//sun				

				MV->pushMatrix();
					glUniform3f(prog->getUniform("ka"), lgt->getKa().x(), lgt->getKa().y(), lgt->getKa().z());
					glUniform3f(prog->getUniform("kd"), lgt->getKd().x(), lgt->getKd().y(), lgt->getKd().z());
					glUniform3f(prog->getUniform("ks"), lgt->getKs().x(), lgt->getKs().y(), lgt->getKs().z());
					glUniform1f(prog->getUniform("s"), lgt->getS());

					MV->translate(lgt->getTransformMesh());
					MV->rotate(lgt->getAngle().x(), xAxis);
					MV->rotate(lgt->getAngle().y(), yAxis);
					MV->rotate(lgt->getAngle().z(), zAxis);
					MV->scale(lgt->getScaleMesh());
					glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					lgt->getShape()->draw(prog);
				MV->popMatrix();

				//bunny 1
				MV->pushMatrix();
					glUniform3f(prog->getUniform("ka"), b1->getKa().x(), b1->getKa().y(), b1->getKa().z());
					glUniform3f(prog->getUniform("kd"), b1->getKd().x(), b1->getKd().y(), b1->getKd().z());
					glUniform3f(prog->getUniform("ks"), b1->getKs().x(), b1->getKs().y(), b1->getKs().z());
					glUniform1f(prog->getUniform("s"), b1->getS());

					MV->translate(b1->getTransformMesh());
					MV->rotate(b1->getAngle().x(), xAxis);
					MV->rotate(b1->getAngle().y(), yAxis);
					MV->rotate(b1->getAngle().z(), zAxis);
					MV->scale(b1->getScaleMesh());
					glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
					b1->getShape()->draw(prog);
				MV->popMatrix();


				////building



				////target 0		
				//if ((targ. <= b.maxX && targ.maxX >= b.minX) &&
				//	(targ.minY <= b.maxY && targ.maxY >= b.minY) &&
				//	(targ.minZ <= b.maxZ && targ.maxZ >= b.minZ)){
					MV->pushMatrix();
						glUniform3f(prog->getUniform("ka"), targ->getKa().x(), targ->getKa().y(), targ->getKa().z());
						glUniform3f(prog->getUniform("kd"), targ->getKd().x(), targ->getKd().y(), targ->getKd().z());
						glUniform3f(prog->getUniform("ks"), targ->getKs().x(), targ->getKs().y(), targ->getKs().z());
						glUniform1f(prog->getUniform("s"), targ->getS());

						MV->translate(Vector3f(0.0f, 0.0f, 70.0f));
						MV->scale(targ->getScaleMesh());
						glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
						targ->getShape()->draw(prog);
					MV->popMatrix();	
				//}	
				////target function 
				


				MV->pushMatrix();
				glUniform3f(prog->getUniform("ka"), targ->getKa().x(), targ->getKa().y(), targ->getKa().z());
				glUniform3f(prog->getUniform("kd"), targ->getKd().x(), targ->getKd().y(), targ->getKd().z());
				glUniform3f(prog->getUniform("ks"), targ->getKs().x(), targ->getKs().y(), targ->getKs().z());
				glUniform1f(prog->getUniform("s"), targ->getS());

				MV->translate(Vector3f(-0.5f, 0.0f, 90.0f));
				MV->scale(targ->getScaleMesh());
				glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
				targ->getShape()->draw(prog);
				MV->popMatrix();				

				////target 2
				MV->pushMatrix();
				glUniform3f(prog->getUniform("ka"), targ->getKa().x(), targ->getKa().y(), targ->getKa().z());
				glUniform3f(prog->getUniform("kd"), targ->getKd().x(), targ->getKd().y(), targ->getKd().z());
				glUniform3f(prog->getUniform("ks"), targ->getKs().x(), targ->getKs().y(), targ->getKs().z());
				glUniform1f(prog->getUniform("s"), targ->getS());

				MV->translate(Vector3f(-0.5f, 0.5f, 80.0f));
				MV->scale(targ->getScaleMesh());
				glUniformMatrix4fv(prog->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
				targ->getShape()->draw(prog);
				MV->popMatrix();


				text->unbind();
			prog->unbind();


			////marble	

			marb->bind();			
				
				MV->pushMatrix(); 
					glEnable(GL_BLEND);
					glDepthMask(GL_FALSE);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					texture0->bind(marb->getUniform("texture0"));
						glUniformMatrix4fv(marb->getUniform("P"), 1, GL_FALSE, P->topMatrix().data());
						glUniformMatrix4fv(marb->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
						glUniform2f(marb->getUniform("screenSize"), (float)width, (float)height);
						
						Particle::draw(marble, marb);
					texture0->unbind();
				MV->popMatrix();	
				glDepthMask(GL_TRUE);
				glDisable(GL_BLEND);
			marb->bind();

		MV->popMatrix();

		//move camera forwards
		fl->pos = fl->pos + Vector3f(0.0f, 0.0f, -0.05f);

	P->popMatrix();
	
	GLSL::checkError(GET_FILE_LINE);
}

void stepParticles()
{
		// This can be parallelized!
		marble[0]->step(t, h, grav, keyToggles);
		t += h;
}

int main(int argc, char **argv)
{
	if(argc < 2) {
		cout << "Please specify the resource directory." << endl;
		return 0;
	}
	RESOURCE_DIR = argv[1] + string("/");

	// Set error callback.
	glfwSetErrorCallback(error_callback);
	// Initialize the library.
	if(!glfwInit()) {
		return -1;
	}
	// Create a windowed mode window and its OpenGL context.
	window = glfwCreateWindow(640, 480, "Jeffrey Tang", NULL, NULL);
	if(!window) {
		glfwTerminate();
		return -1;
	}
	// Make the window's context current.
	glfwMakeContextCurrent(window);
	// Initialize GLEW.
	glewExperimental = true;
	if(glewInit() != GLEW_OK) {
		cerr << "Failed to initialize GLEW" << endl;
		return -1;
	}
	glGetError(); // A bug in glewInit() causes an error that we can safely ignore.
	cout << "OpenGL version: " << glGetString(GL_VERSION) << endl;
	cout << "GLSL version: " << glGetString(GL_SHADING_LANGUAGE_VERSION) << endl;
	GLSL::checkVersion();
	// Set vsync.
	glfwSwapInterval(1);
	// Set keyboard callback.
	glfwSetKeyCallback(window, key_callback);
	// Set char callback.
	glfwSetCharCallback(window, char_callback);
	// Set cursor position callback.
	glfwSetCursorPosCallback(window, cursor_position_callback);
	// Set mouse button callback.
	glfwSetMouseButtonCallback(window, mouse_button_callback);
	// Set the window resize call back.
	glfwSetFramebufferSizeCallback(window, resize_callback);
	// Initialize scene.
	init();
	// Loop until the user closes the window.
	while(!glfwWindowShouldClose(window)) {
		// Render scene.
		if (keyToggles[(unsigned int)'p'] == true)
			render(); 
		stepParticles();
		// Swap front and back buffers.
		glfwSwapBuffers(window);
		// Poll for and process events.
		glfwPollEvents();
	}
	// Quit program.
	glfwDestroyWindow(window);
	glfwTerminate();
	return 0;
}
